
local addon_name, addon = ...

local module = addon:NewModule("Locations", "AceEvent-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("PetLeash")
local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

local _G = _G
local assert, ipairs, pairs, rawget, wipe = _G.assert, _G.ipairs, _G.pairs, _G.rawget, _G.wipe
local tonumber, tostring = _G.tonumber, _G.tostring

--
--
--

local options = {
    type = "group",
    name = L["Locations"],
    order = 11,
    cmdHidden = true,
    args = {
        specialLocations = {
            type = "group",
            name = L["Special Locations"],
            order = 1,
            args = {
                description = {
                    type = "description",
                    name = L["Special Locations are predefined areas that cover a certain type of zone."]
                }
            },
            plugins = { data = {} }
        },
        customLocations = {
            type = "group",
            name = L["Custom Locations"],
            order = 2,
            args = {
               addCurrentZone = {
                    type = "execute",
                    name = L["Add Current Zone"],
                    order = 1,
                    func = function(info) 
                        module:AddCustomLocation(GetZoneText())
                    end,
                },
                addCurrentSubZone = {
                    type = "execute",
                    name = L["Add Current Subzone"],
                    order = 1,
                    func = function(info) 
                        module:AddCustomLocation(GetSubZoneText())
                    end,
                },
                --addNamedZone = {
                -- 
                --}   
            },
            plugins = { data = {} }
        },
    }
}

local defaults = {
    profile = {
        sets = {
            -- locations
            customLocations = {
                -- custom locations
                ["*"] = {
                    enable = false,
                    immediate = true,
                    inherit = false,
                    pets = {}           -- {petid, petid, ...}
                }
            },
            specialLocations = {
                -- premade (special) locations
                ["*"] = {
                    enable = false,
                    immediate = true,
                    inherit = false,
                    pets = {}           -- {petid, petid, ...}
                }
            }
        }
    }
}

--
--
--

local special_locations = {
    city = {
        name = L["City"],
        func = function() return IsResting() end,
    },
    battleground = {
        name = BATTLEGROUND,
        func = function() return addon.InBattlegroundOrArena() end,
    },
    instance = {
        name = L["Instance"],
        func = function() return addon.InInstanceOrRaid() end,
    },
}

local LOCATION_TYPES = {customLocations=1, specialLocations=1}

local config_getLocationArgs, buildConfigLocation

function module:OnInitialize()
    self.db = addon.db:RegisterNamespace("Locations", defaults)
    
    self:RegisterEvent("ZONE_CHANGED")
    self:RegisterEvent("ZONE_CHANGED_INDOORS")
    self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
    self:RegisterEvent("PLAYER_UPDATE_RESTING")
end

function module:OnEnable()
    self:TryInitLocation()
end

function module:SetupOptions(setup)
    AceConfig:RegisterOptionsTable(self.name, options)
    self:UpdateConfigTables()
    setup(self.name, L["Locations"])
end

function module:AddCustomLocation(name)
    if(not name or name == "") then
        return
    end
    
    self.db.profile.sets.customLocations[name].enable = true      -- touch
    
    self:UpdateCustomLocationConfigTables()
end

function module:DeleteCustomLocation(name)
    wipe(self.db.profile.sets.customLocations[name])
    self.db.profile.sets.customLocations[name] = nil
    self:UpdateCustomLocationConfigTables()
    self:DoLocationCheck(false)
end

function module:GetLocationPet(ltype, name, petid)
    assert(LOCATION_TYPES[ltype])
    
    for i,v in ipairs(self.db.profile.sets[ltype][name].pets) do
        if(petid == v) then
            return i
        end
    end
end

function module:SetLocationPet(ltype, name, petid, value)
    assert(LOCATION_TYPES[ltype])

    local t = self.db.profile.sets[ltype][name].pets
    local iszit = self:GetLocationPet(ltype, name, petid)
    if(value and not iszit) then
        table.insert(t, petid)
    elseif(not value and iszit) then
        table.remove(t, iszit)
    end
    
    self:DoLocationCheck(false)
end

--
-- config
--

function module:UpdateConfigTables()
    self:UpdateCustomLocationConfigTables(true)
    self:UpdateSpecialLocationConfigTables(true)
end

-- dirty bits for updating custom locations

local function config_location_pettoggle_get(info)
    return module:GetLocationPet(info[#info-3], info[#info-2], tonumber(info[#info]))
end

local function config_location_pettoggle_set(info, val)
    module:SetLocationPet(info[#info-3], info[#info-2], tonumber(info[#info]), val)
end

local function config_location_delete(info)
    module:DeleteCustomLocation(info[#info-1])
end

local function config_location_enable_set(info,v)
    module.db.profile.sets[info[#info-2]][info[#info-1]].enable = v
    module:DoLocationCheck(false)
end

local function config_location_enable_get(info)
    assert(LOCATION_TYPES[info[#info-2]])
    return module.db.profile.sets[info[#info-2]][info[#info-1]].enable
end

local function config_location_immediate_set(info,v)
    assert(LOCATION_TYPES[info[#info-2]])
    module.db.profile.sets[info[#info-2]][info[#info-1]].immediate=v
end

local function config_location_immediate_get(info)
    assert(LOCATION_TYPES[info[#info-2]])
    return module.db.profile.sets[info[#info-2]][info[#info-1]].immediate
end

local function config_location_inherit_set(info, v)
    assert(LOCATION_TYPES[info[#info-2]])
    local loc = module.db.profile.sets[info[#info-2]][info[#info-1]]
    if(not v) then
        loc.inherit = false
        module:UpdateConfigTables()
    elseif(v and not loc.inherit) then
        -- only set if we're not going to clobber it
        -- and not ourselves
        loc.inherit = true
        module:UpdateConfigTables()
    end
    
    module:DoLocationCheck(false)
end

local function config_location_inherit_get(info)
    assert(LOCATION_TYPES[info[#info-2]])
    return module.db.profile.sets[info[#info-2]][info[#info-1]].inherit
end

local loc_pet_config = {
    type = "group",
    name = "",
    order = 10,
    args = {},
    inline = true
}
local loc_inherit_config = {
    type = "select",
    name = L["Inherits From"],
    order = 11,
    values = {},
    get = function(info)
        assert(LOCATION_TYPES[info[#info-2]])
        
        local inherit = module.db.profile.sets[info[#info-2]][info[#info-1]].inherit
        if(inherit ~= true) then
            return inherit
        end
    end,
    set = function(info, val)
        assert(LOCATION_TYPES[info[#info-2]])
        module.db.profile.sets[info[#info-2]][info[#info-1]].inherit = val
        module:DoLocationCheck(false)
    end
}

function module:UpdateCustomLocationConfigTables(nosignal)
    local pet_args = loc_pet_config.args
    wipe(pet_args)
    
    for i,petinfo in ipairs(addon.PetInfo.Pets) do
        if petinfo.isOwned and petinfo.isSummonable then
            local name = petinfo.customName or petinfo.name
            pet_args[tostring(petinfo.petID)] = {
                type = "toggle",
                name = name,
                get = config_location_pettoggle_get,
                set = config_location_pettoggle_set
            }
        end
    end
    
    local loc_args = options.args.customLocations.plugins.data
    wipe(loc_args)          -- TODO: check to see if locations is dirty before wiping
    wipe(loc_inherit_config.values)
    
    for name,data in pairs(self.db.profile.sets.customLocations) do
        if(data.enable) then
            if(not loc_inherit_config.values[name]) then
                loc_inherit_config.values[name] = name
            end
        
            buildConfigLocation(loc_args,
                                name,
                                name,
                                self.db.profile.sets.customLocations[name].inherit,
                                "customLocations")
        end
    end
    
    if(not nosignal) then
        AceConfigRegistry:NotifyChange("PetLeash")
    end
end

function module:UpdateSpecialLocationConfigTables(nosignal)
    for key,data in pairs(special_locations) do
        buildConfigLocation(
                options.args.specialLocations.plugins.data,
                key,
                data.name,
                self.db.profile.sets.specialLocations[key].inherit,
                "specialLocations")
    end
    
    if(not nosignal) then
        AceConfigRegistry:NotifyChange("PetLeash")
    end
end

function buildConfigLocation(args, key, name, inherit, ctype)
    if(not args[key]) then
        args[key] = config_getLocationArgs(name, ctype)
    end
    
    if(inherit) then
        args[key].args.pets = nil
        args[key].args.inherits = loc_inherit_config
    else
        args[key].args.pets = loc_pet_config
        args[key].args.inherits = nil
    end
end

function config_getLocationArgs(name, ctype)
    local deleteMe, enableMe

    if(ctype == "customLocations") then
        deleteMe = {
            type = "execute",
            name = DELETE,
            order = 1,
            func = config_location_delete
        }
    elseif(ctype == "specialLocations") then
        deleteMe = {
            type = "toggle",
            name = ENABLE,
            order = 1,
            get = config_location_enable_get,
            set = config_location_enable_set,
        }
    end

    return {
        type = "group",
        name = name,
        args = {
            deleteMe = deleteMe,
            enableMe = enableMe,
            immediate = {
                type = "toggle",
                name = L["Immediate"],
                desc = L["Immediately switch pets upon zone change."],
                order = 2,
                set = config_location_immediate_set,
                get = config_location_immediate_get
            },
            inherit = {
                type = "toggle",
                name = L["Inherits"],
                desc = L["Use a pet list from another location."],
                order = 2,
                set = config_location_inherit_set,
                get = config_location_inherit_get,
            },
            seperator = {
                type = "header",
                name = "",
                order = 3,
            },
        }
    }
end

--
-- switcher code
--

function module:ZONE_CHANGED()
    self:DoLocationCheck(true)
end

function module:ZONE_CHANGED_INDOORS()
    self:DoLocationCheck(true)
end

function module:ZONE_CHANGED_NEW_AREA()
    self:DoLocationCheck(true)
end

function module:PLAYER_UPDATE_RESTING()
    self:DoLocationCheck(true)
end

function module:TryInitLocation()
    if(GetZoneText() == "" or GetZoneText() == nil) then
        return
    end
    
    self:UpdateConfigTables()
    self:DoLocationCheck(false)
    
    self.TryInitLocation = function() end
end

local checkZone

function module:DoLocationCheck(allow_immediate)
    local cur_zone = GetZoneText()
    local cur_subzone = GetSubZoneText()
    local cur_pet = addon:HasPet()

    -- custom zone checks
    if(checkZone(self, "customLocations", cur_subzone, cur_pet, allow_immediate)) then
        return
    end
    if(checkZone(self, "customLocations", cur_zone, cur_pet, allow_immediate)) then
        return
    end

    -- special zone checks
    for key,data in pairs(special_locations) do
        if(data.func()) then
            if(checkZone(self, "specialLocations", key, cur_pet, allow_immediate)) then
                return
            end
        end
    end
   
    -- nothing doing
    self.override_pets = nil
end

function checkZone(self, ltype, zonename, curpet, allow_immediate)
    if(not zonename or zonename == "") then
        return
    end

    -- don't let AceDB generate an entry for us
    local locdata = rawget(self.db.profile.sets[ltype], zonename)

    -- make sure entry exists
    if(not locdata or not locdata.enable) then
        return
    end
    
    local pets = locdata.pets
    if(locdata.inherit and locdata.inherit ~= true
                and self.db.profile.sets.customLocations[locdata.inherit]) then
        pets = self.db.profile.sets.customLocations[locdata.inherit].pets
    end
    
    if pets and #pets > 0 then
        self.override_pets = pets
        if(allow_immediate and locdata.immediate) then
            local i
            for id,petid in pairs(pets) do
                if petid == curpet then
                    i = id
                    break
                end
            end
            
            if(not i) then
                -- try to resummon now
                addon:ResummonPet()
            end
        end
        
        return true
    end
end

function module:PickPet()
    if self.override_pets and #self.override_pets > 0 then
        return addon:PickPetFromList(self.override_pets)
    end
end
